@@ -1,3 +1,4 @@
+#!/usr/bin/perl
 # -*- perl -*-
 #
 #   DBD::File - A base class for implementing DBI drivers that
@@ -9,7 +10,7 @@
 #
 #  The original author is Jochen Wiedmann.
 #
-#  Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
+#  Copyright (C) 2009-2020 by H.Merijn Brand & Jens Rehsack
 #  Copyright (C) 2004 by Jeff Zucker
 #  Copyright (C) 1998 by Jochen Wiedmann
 #
@@ -33,14 +34,12 @@ use warnings;
 
 use base qw( DBI::DBD::SqlEngine );
 use Carp;
-use vars qw( @ISA $VERSION $drh );
 
-$VERSION = "0.44";
+our $VERSION = "0.44";
 
-$drh = undef;		# holds driver handle(s) once initialized
+our $drh = undef;		# holds driver handle(s) once initialized
 
-sub driver ($;$)
-{
+sub driver ($;$) {
     my ($class, $attr) = @_;
 
     # Drivers typically use a singleton object for the $drh
@@ -71,8 +70,7 @@ sub driver ($;$)
     return $drh->{$class};
     } # driver
 
-sub CLONE
-{
+sub CLONE {
     undef $drh;
     } # CLONE
 
@@ -83,15 +81,12 @@ package DBD::File::dr;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
-
 use Carp;
 
-@DBD::File::dr::ISA           = qw( DBI::DBD::SqlEngine::dr );
-$DBD::File::dr::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::dr );
+our $imp_data_size = 0;
 
-sub dsn_quote
-{
+sub dsn_quote {
     my $str = shift;
     ref     $str and return "";
     defined $str or  return "";
@@ -102,37 +97,49 @@ sub dsn_quote
 # XXX rewrite using TableConfig ...
 sub default_table_source { "DBD::File::TableSource::FileSystem" }
 
-sub connect
-{
+sub connect {
     my ($drh, $dbname, $user, $auth, $attr) = @_;
 
     # We do not (yet) care about conflicting attributes here
     # my $dbh = DBI->connect ("dbi:CSV:f_dir=test", undef, undef, { f_dir => "text" });
     # will test here that both test and text should exist
-    if (my $attr_hash = (DBI->parse_dsn ($dbname))[3]) {
-	if (defined $attr_hash->{f_dir} && ! -d $attr_hash->{f_dir}) {
-	    my $msg = "No such directory '$attr_hash->{f_dir}";
-	    $drh->set_err (2, $msg);
-	    $attr_hash->{RaiseError} and croak $msg;
-	    return;
+    #
+    # Parsing on our own similar to parse_dsn to find attributes in 'dbname' parameter.
+    if ($dbname) {
+	my $attr_hash = {
+	    map { (m/^\s* (\S+) \s*(?: =>? | , )\s* (\S*) \s*$/x) }
+	    split m/;/ => $dbname };
+	if (defined $attr_hash->{f_dir}) {
+	    my $f_dir = $attr_hash->{f_dir};
+	    # DSN escapes the : in Windows' path, which is not accepted by -d
+	    #   D\\:\\\\Test\\\\DBI-01\\\\test_output_12345
+	    # ->  D:\\\\Test\\\\DBI-01\\\\test_output_12345
+	    $^O eq "MSWin32" and $f_dir =~ s{^([a-zA-Z])\\+:}{$1:};
+	    unless (-d $f_dir) {
+		my $msg = "No such directory '$attr_hash->{f_dir}";
+		$drh->set_err (2, $msg);
+		$attr_hash->{RaiseError} and croak $msg;
+		return;
+		}
 	    }
 	}
-    if ($attr and defined $attr->{f_dir} && ! -d $attr->{f_dir}) {
-	my $msg = "No such directory '$attr->{f_dir}";
-	$drh->set_err (2, $msg);
-	$attr->{RaiseError} and croak $msg;
-	return;
+    if ($attr and defined $attr->{f_dir}) {
+	my $f_dir = $attr->{f_dir};
+	$^O eq "MSWin32" and $f_dir =~ s{^([a-zA-Z])\\+:}{$1:};
+	unless (-d $f_dir) {
+	    my $msg = "No such directory '$attr->{f_dir}";
+	    $drh->set_err (2, $msg);
+	    return;
+	    }
 	}
 
     return $drh->SUPER::connect ($dbname, $user, $auth, $attr);
     } # connect
 
-sub disconnect_all
-{
+sub disconnect_all {
     } # disconnect_all
 
-sub DESTROY
-{
+sub DESTROY {
     undef;
     } # DESTROY
 
@@ -143,18 +150,15 @@ package DBD::File::db;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
-
 use Carp;
 require File::Spec;
 require Cwd;
 use Scalar::Util qw( refaddr ); # in CORE since 5.7.3
 
-@DBD::File::db::ISA           = qw( DBI::DBD::SqlEngine::db );
-$DBD::File::db::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::db );
+our $imp_data_size = 0;
 
-sub data_sources
-{
+sub data_sources {
     my ($dbh, $attr, @other) = @_;
     ref ($attr) eq "HASH" or $attr = {};
     exists $attr->{f_dir}        or $attr->{f_dir}        = $dbh->{f_dir};
@@ -162,16 +166,14 @@ sub data_sources
     return $dbh->SUPER::data_sources ($attr, @other);
     } # data_source
 
-sub set_versions
-{
+sub set_versions {
     my $dbh = shift;
     $dbh->{f_version} = $DBD::File::VERSION;
 
     return $dbh->SUPER::set_versions ();
     } # set_versions
 
-sub init_valid_attributes
-{
+sub init_valid_attributes {
     my $dbh = shift;
 
     $dbh->{f_valid_attrs} = {
@@ -195,8 +197,7 @@ sub init_valid_attributes
     return $dbh->SUPER::init_valid_attributes ();
     } # init_valid_attributes
 
-sub init_default_attributes
-{
+sub init_default_attributes {
     my ($dbh, $phase) = @_;
 
     # must be done first, because setting flags implicitly calls $dbdname::db->STORE
@@ -232,8 +233,7 @@ sub init_default_attributes
     return $dbh;
     } # init_default_attributes
 
-sub validate_FETCH_attr
-{
+sub validate_FETCH_attr {
     my ($dbh, $attrib) = @_;
 
     $attrib eq "f_meta" and $dbh->{sql_engine_in_gofer} and $attrib = "sql_meta";
@@ -241,8 +241,7 @@ sub validate_FETCH_attr
     return $dbh->SUPER::validate_FETCH_attr ($attrib);
     } # validate_FETCH_attr
 
-sub validate_STORE_attr
-{
+sub validate_STORE_attr {
     my ($dbh, $attrib, $value) = @_;
 
     if ($attrib eq "f_dir" && defined $value) {
@@ -262,8 +261,7 @@ sub validate_STORE_attr
     return $dbh->SUPER::validate_STORE_attr ($attrib, $value);
     } # validate_STORE_attr
 
-sub get_f_versions
-{
+sub get_f_versions {
     my ($dbh, $table) = @_;
 
     my $class = $dbh->{ImplementorClass};
@@ -297,10 +295,8 @@ package DBD::File::st;
 use strict;
 use warnings;
 
-use vars qw( @ISA $imp_data_size );
-
-@DBD::File::st::ISA           = qw( DBI::DBD::SqlEngine::st );
-$DBD::File::st::imp_data_size = 0;
+our @ISA           = qw( DBI::DBD::SqlEngine::st );
+our $imp_data_size = 0;
 
 my %supported_attrs = (
     TYPE      => 1,
@@ -308,8 +304,7 @@ my %supported_attrs = (
     NULLABLE  => 1,
     );
 
-sub FETCH
-{
+sub FETCH {
     my ($sth, $attr) = @_;
 
     if ($supported_attrs{$attr}) {
@@ -392,10 +387,9 @@ use warnings;
 
 use IO::Dir;
 
-@DBD::File::TableSource::FileSystem::ISA = "DBI::DBD::SqlEngine::TableSource";
+our @ISA = "DBI::DBD::SqlEngine::TableSource";
 
-sub data_sources
-{
+sub data_sources {
     my ($class, $drh, $attr) = @_;
     my $dir = $attr && exists $attr->{f_dir}
 	? $attr->{f_dir}
@@ -434,8 +428,7 @@ sub data_sources
     return @dsns;
     } # data_sources
 
-sub avail_tables
-{
+sub avail_tables {
     my ($self, $dbh) = @_;
 
     my $dir = $dbh->{f_dir};
@@ -483,7 +476,7 @@ use warnings;
 
 use Carp;
 
-@DBD::File::DataSource::Stream::ISA = "DBI::DBD::SqlEngine::DataSource";
+our @ISA = "DBI::DBD::SqlEngine::DataSource";
 
 # We may have a working flock () built-in but that doesn't mean that locking
 # will work on NFS (flock () may hang hard)
@@ -496,8 +489,7 @@ my $locking = eval {
     1;
     };
 
-sub complete_table_name
-{
+sub complete_table_name {
     my ($self, $meta, $file, $respect_case) = @_;
 
     my $tbl = $file;
@@ -517,8 +509,7 @@ sub complete_table_name
     return $tbl;
     } # complete_table_name
 
-sub apply_encoding
-{
+sub apply_encoding {
     my ($self, $meta, $fn) = @_;
     defined $fn or $fn = "file handle " . fileno ($meta->{fh});
     if (my $enc = $meta->{f_encoding}) {
@@ -530,8 +521,7 @@ sub apply_encoding
 	}
     } # apply_encoding
 
-sub open_data
-{
+sub open_data {
     my ($self, $meta, $attrs, $flags) = @_;
 
     $flags->{dropMode} and croak "Can't drop a table in stream";
@@ -572,14 +562,13 @@ package DBD::File::DataSource::File;
 use strict;
 use warnings;
 
-@DBD::File::DataSource::File::ISA = "DBD::File::DataSource::Stream";
+our @ISA = "DBD::File::DataSource::Stream";
 
 use Carp;
 
 my $fn_any_ext_regex = qr/\.[^.]*/;
 
-sub complete_table_name
-{
+sub complete_table_name {
     my ($self, $meta, $file, $respect_case, $file_is_table) = @_;
 
     $file eq "." || $file eq ".."	and return; # XXX would break a possible DBD::Dir
@@ -593,7 +582,7 @@ sub complete_table_name
 	    }
 	}
 
-    # (my $tbl = $file) =~ s/$ext$//i;
+    # (my $tbl = $file) =~ s/\Q$ext\E$//i;
     my ($tbl, $basename, $dir, $fn_ext, $user_spec_file, $searchdir);
     if ($file_is_table and defined $meta->{f_file}) {
 	$tbl = $file;
@@ -602,7 +591,7 @@ sub complete_table_name
 	$user_spec_file = 1;
 	}
     else {
-	($basename, $dir, undef) = File::Basename::fileparse ($file, $ext);
+	($basename, $dir, undef) = File::Basename::fileparse ($file, qr{\Q$ext\E});
 	# $dir is returned with trailing (back)slash. We just need to check
 	# if it is ".", "./", or ".\" or "[]" (VMS)
 	if ($dir =~ m{^(?:[.][/\\]?|\[\])$} && ref $meta->{f_dir_search} eq "ARRAY") {
@@ -673,12 +662,12 @@ sub complete_table_name
 	    }
 	@f > 0 && @f <= 2 and $file = $f[0];
 	!$respect_case && $meta->{sql_identifier_case} == 4 and # XXX SQL_IC_MIXED
-	    ($tbl = $file) =~ s/$ext$//i;
+	    ($tbl = $file) =~ s/\Q$ext\E$//i;
 
 	my $tmpfn = $file;
 	if ($ext && $req) {
             # File extension required
-            $tmpfn =~ s/$ext$//i or return;
+            $tmpfn =~ s/\Q$ext\E$//i or return;
             }
 	}
 
@@ -696,8 +685,7 @@ sub complete_table_name
     return $tbl;
     } # complete_table_name
 
-sub open_data
-{
+sub open_data {
     my ($self, $meta, $attrs, $flags) = @_;
 
     defined $meta->{f_fqfn} && $meta->{f_fqfn} ne "" or croak "No filename given";
@@ -765,7 +753,7 @@ package DBD::File::Statement;
 use strict;
 use warnings;
 
-@DBD::File::Statement::ISA = qw( DBI::DBD::SqlEngine::Statement );
+our @ISA = qw( DBI::DBD::SqlEngine::Statement );
 
 # ====== SQL::TABLE ============================================================
 
@@ -781,7 +769,7 @@ require File::Spec;
 require Cwd;
 require Scalar::Util;
 
-@DBD::File::Table::ISA = qw( DBI::DBD::SqlEngine::Table );
+our @ISA = qw( DBI::DBD::SqlEngine::Table );
 
 # ====== UTILITIES ============================================================
 
@@ -829,15 +817,13 @@ else {
 # The functions file2table, init_table_meta, default_table_meta and
 # get_table_meta are using $self arguments for polymorphism only. The
 # must not rely on an instantiated DBD::File::Table
-sub file2table
-{
+sub file2table {
     my ($self, $meta, $file, $file_is_table, $respect_case) = @_;
 
     return $meta->{sql_data_source}->complete_table_name ($meta, $file, $respect_case, $file_is_table);
     } # file2table
 
-sub bootstrap_table_meta
-{
+sub bootstrap_table_meta {
     my ($self, $dbh, $meta, $table, @other) = @_;
 
     $self->SUPER::bootstrap_table_meta ($dbh, $meta, $table, @other);
@@ -859,8 +845,7 @@ sub bootstrap_table_meta
 				 : "DBD::File::DataSource::File";
     } # bootstrap_table_meta
 
-sub get_table_meta ($$$$;$)
-{
+sub get_table_meta ($$$$;$) {
     my ($self, $dbh, $table, $file_is_table, $respect_case) = @_;
 
     my $meta = $self->SUPER::get_table_meta ($dbh, $table, $respect_case, $file_is_table);
@@ -887,15 +872,13 @@ __PACKAGE__->register_compat_map (\%compat_map);
 # ====== DBD::File <= 0.40 compat stuff ========================================
 
 # compat to 0.38 .. 0.40 API
-sub open_file
-{
+sub open_file {
     my ($className, $meta, $attrs, $flags) = @_;
 
     return $className->SUPER::open_data ($meta, $attrs, $flags);
     } # open_file
 
-sub open_data
-{
+sub open_data {
     my ($className, $meta, $attrs, $flags) = @_;
 
     # compat to 0.38 .. 0.40 API
@@ -908,8 +891,7 @@ sub open_data
 
 # ====== SQL::Eval API =========================================================
 
-sub drop ($)
-{
+sub drop ($) {
     my ($self, $data) = @_;
     my $meta = $self->{meta};
     # We have to close the file before unlinking it: Some OS'es will
@@ -924,8 +906,7 @@ sub drop ($)
     return 1;
     } # drop
 
-sub seek ($$$$)
-{
+sub seek ($$$$) {
     my ($self, $data, $pos, $whence) = @_;
     my $meta = $self->{meta};
     if ($whence == 0 && $pos == 0) {
@@ -939,8 +920,7 @@ sub seek ($$$$)
 	croak "Error while seeking in " . $meta->{f_fqfn} . ": $!";
     } # seek
 
-sub truncate ($$)
-{
+sub truncate ($$) {
     my ($self, $data) = @_;
     my $meta = $self->{meta};
     $meta->{fh}->truncate ($meta->{fh}->tell ()) or
@@ -948,8 +928,7 @@ sub truncate ($$)
     return 1;
     } # truncate
 
-sub DESTROY
-{
+sub DESTROY {
     my $self = shift;
     my $meta = $self->{meta};
     $meta->{fh} and $meta->{fh}->close ();
@@ -1033,13 +1012,17 @@ C<< $sth->execute >>; undef for non-select statements.
 
 =head3 Unsupported DBI attributes and methods
 
-=head4 bind_param_inout
+=over 2
 
-=head4 CursorName
+=item bind_param_inout
 
-=head4 LongReadLen
+=item CursorName
 
-=head4 LongTruncOk
+=item LongReadLen
+
+=item LongTruncOk
+
+=back
 
 =head3 DBD::File specific attributes
 
@@ -1058,6 +1041,11 @@ directory) when the dbh attribute is set.
 
   f_dir => "/data/foo/csv",
 
+If C<f_dir> is set to a non-existing location, the connection will fail.
+See CVE-2014-10401 for reasoning. Because of this, folders to use cannot
+be created after the connection, but must exist before the connection is
+initiated.
+
 See L<KNOWN BUGS AND LIMITATIONS>.
 
 =head4 f_dir_search
@@ -1225,12 +1213,12 @@ releases on
 
   @ary = DBI->data_sources ($driver);
   @ary = DBI->data_sources ($driver, \%attr);
-  
+
   @ary = $dbh->data_sources ();
   @ary = $dbh->data_sources (\%attr);
 
   @names = $dbh->tables ($catalog, $schema, $table, $type);
-  
+
   $sth = $dbh->table_info ($catalog, $schema, $table, $type);
   $sth = $dbh->table_info ($catalog, $schema, $table, $type, \%attr);
 
@@ -1430,7 +1418,7 @@ The original author is Jochen Wiedmann.
 
 =head1 COPYRIGHT AND LICENSE
 
- Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
+ Copyright (C) 2009-2020 by H.Merijn Brand & Jens Rehsack
  Copyright (C) 2004-2009 by Jeff Zucker
  Copyright (C) 1998-2004 by Jochen Wiedmann
 
